(* $Id: godi_file_helper.ml 1072 2012-07-16 12:30:50Z gerd $  
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
*)

open Godi_win32

let symlink_buf = String.create 4096
  (* impl restriction: symlink shotcuts must fit into 4096 bytes *)

(* Cygwin-1.5 symlinks *)
let symlink_magic =
  "\x4c\x00\x00\x00\
   \x01\x14\x02\x00\
   \x00\x00\x00\x00\
   \xc0\x00\x00\x00\
   \x00\x00\x00\x46"

(* Cygwin-1.7 symlinks *)
let symlink_magic_17 =
  "!<symlink>\xff\xfe"
    (* maybe the FF FE is a BOM *)

let short s p =
  if p + 1 >= String.length s then raise Not_found;
  (Char.code s.[p]) lor ((Char.code s.[p+1]) lsl 8)


let lnk_re = Str.regexp ".*\\.[lL][nN][kK]$"

let read_shortcut_symlink path =
  (* Return [Some target] when [path] is a Cygwin symlink, or [None] otherwise.
     [path] must be a Win32 path.

     We only support new-style symlinks, i.e. the encoding as Windows
     shortcut.
   *)
  
  try
    if not(Str.string_match lnk_re path 0) then
      raise Not_found;

    let st = Unix.stat path in
    if st.Unix.st_kind <> Unix.S_REG then raise Not_found;
    if st.Unix.st_perm land 0o200 <> 0 then raise Not_found;

    let f = open_in_bin path in
    ( try
	let l = in_channel_length f in
	if l < 80 then raise Not_found;
	let l = min l (String.length symlink_buf) in
	really_input f symlink_buf 0 l;
	close_in f;

	let magic = String.sub symlink_buf 0 20 in
	if magic <> symlink_magic then raise Not_found;

	let flags = Char.code symlink_buf.[20] in
	if flags <> 12 && flags <> 13 then raise Not_found;
	let have_idlist = (flags = 13) in

	if String.sub symlink_buf 21 3 <> "\x00\x00\x00" then 
	  raise Not_found;
	if String.sub symlink_buf 60 4 <> "\x01\x00\x00\x00" then 
	  raise Not_found;

	let p = ref 76 in  (* after header *)
	if have_idlist then (
	  let idlist_len = short symlink_buf 76 in
	  p := !p + 2 + idlist_len;
	);

	let descr_len = short symlink_buf !p in
	let descr_start = !p + 2 in
	if descr_start + descr_len > l then raise Not_found;

	let descr = String.sub symlink_buf descr_start descr_len in
	p := descr_start + descr_len;

	let relpath_len = short symlink_buf !p in
	p := !p + 2 + relpath_len;

	if !p > l then raise Not_found;

	if !p = l then
	  Some descr
	else
	  if l < 4096 then
	    Some(String.sub symlink_buf !p (l - !p))
	  else
	    raise Not_found (* Well, we don't know... *)
      with
	| error ->
	    close_in f;
	    raise error
    )
  with
    | Unix.Unix_error(_,_,_)
    | Sys_error _
    | Not_found ->
	None


let read_cyg17_symlink path =
  (* Return [Some target] when [path] is a Cygwin-1.7 symlink, or [None] 
     otherwise.
     [path] must be a Win32 path.
   *)
  try
    let st = Unix.stat path in
    if st.Unix.st_kind <> Unix.S_REG then raise Not_found;
    if not(Godi_win32.get_file_system_attribute path) then raise Not_found;

    let f = open_in_bin path in
    ( try
	let l = in_channel_length f in
	if l < 14 then raise Not_found;
	let l = min l (String.length symlink_buf) in
	really_input f symlink_buf 0 l;
	close_in f;

	let magic = String.sub symlink_buf 0 12 in
	if magic <> symlink_magic_17 then raise Not_found;

        if symlink_buf.[l-2] <> '\000' || symlink_buf.[l-1] <> '\000' then
          raise Not_found;

        (* Ocaml always uses the ANSI charset. So convert UTF-16LE to
           ANSI *)
        let ansi_enc =
          match Netconversion.user_encoding() with
            | None -> `Enc_usascii
            | Some enc -> enc in
        let s1 = String.sub symlink_buf 12 (l-14) in
        let s2 = 
          Netconversion.convert
            ~subst:(fun _ -> "?")
            ~in_enc:`Enc_utf16_le
            ~out_enc:ansi_enc
            s1 in
        Some s2
      with
	| error ->
	    close_in f;
	    raise error
    )
  with
    | Unix.Unix_error(_,_,_)
    | Sys_error _
    | Not_found ->
	None


let write_cyg17_symlink target path =
  let f = open_out_bin path in
  try
    output_string f symlink_magic_17;
    
    let ansi_enc =
      match Netconversion.user_encoding() with
	| None -> `Enc_usascii
	| Some enc -> enc in
    let target16 = 
      Netconversion.convert
	~in_enc:ansi_enc
	~out_enc:`Enc_utf16_le
	target in
    output_string f target16;
    output_string f "\000\000";
    close_out f;
    Godi_win32.set_file_system_attribute path true
  with
    | error ->
	close_out_noerr f;
	raise error



let read_symlink path =
  match read_cyg17_symlink path with
    | None -> 
        ( match read_shortcut_symlink (path ^ ".lnk") with
           | None -> None
           | Some p -> Some(p, path ^ ".lnk")
        )
    | Some p -> Some(p, path)
        

type mount_flags =
    { (* mount_symlink : bool;   (* 0x0001 *) *)
      mount_binary : bool;    (* 0x0002 *)
      (* mount_system : bool;    (* 0x0008 *) *)
      mount_exec : bool;      (* 0x0010 *)
      mount_cygdrive : bool;  (* 0x0020 *)
      mount_cygwin_exec : bool; (* 0x0040 *)
      (* mount_mixed: bool;      (* 0x0080 *)*)
      mount_notexec : bool;   (* 0x0100 *)
      (* mount_devfs : bool;     (* 0x0200 *) *)
      (* mount_proc : bool;      (* 0x0400 *) *)
      (* mount_enc : bool;       (* 0x0800 *) *)
      (* mount_ro : bool;        (* 0x1000 *) *)

      mount_acl : bool;
      mount_bind : bool;
      mount_dos : bool;
      mount_ihash : bool;
      mount_user : bool;
      mount_posix : bool;
    }

type mount_entry =
    { me_posix : string list;
        (* The posix path as component list, e.g. "/usr/bin" becomes
           [ "usr"; "bin" ]. "", ".", and ".." cannot occur.
           "/" is represented as [].
	 *)
      me_win32 : string option;
        (* The win32 path using slashes, if applicable. (For bind mounts,
           this is the destination path).
         *)
      me_flags : mount_flags;
    }


type mount_table =
    { mt_entries : mount_entry list;
      (* See also comment for get_mount_table *)

      mt_cygdrive_prefix : string;
      mt_cygdrive_flags : mount_flags
    }


(*
let read_mount_flags flag =
  { mount_binary = (flag land 0x0002) <> 0;
    mount_exec = (flag land 0x0010) <> 0;
    mount_cygdrive = (flag land 0x0020) <> 0;
    mount_cygwin_exec = (flag land 0x0040) <> 0;
    mount_notexec = (flag land 0x0100) <> 0;
    mount_acl = false;
    mount_bind = false;
    mount_dos = false;
    mount_ihash = false;
    mount_user = false;
    mount_posix = false
  }

let mount_table_path =
  [ "software"; "cygnus solutions"; "cygwin"; "mounts v2" ]


let rec walk_reg_path root closeflag path =
  match path with
    | [] ->
	root
    | key :: path' ->
	let hkey = reg_open_key root key in
	if closeflag then reg_close_key root;
	walk_reg_path hkey true path'


let reg_string name =
  function
    | RV_SZ s -> s
    | _ -> raise (Registry_error ("String value expected for " ^ name))

let reg_int name =
  function
    | RV_DWORD i -> Int32.to_int i
    | _ -> raise (Registry_error ("DWORD value expected for " ^ name))
 *)


let slash_re = Str.regexp "/"
let backslash_re = Str.regexp "\\\\"

let slashify p =
  Str.global_replace backslash_re "/" p

let backslashify p =
  Str.global_replace slash_re "\\\\" p

let split_posix_path p =
  match Str.split_delim slash_re p with
    | [""; "" ] ->
	[]   (* for / *)
    | "" :: p' ->
	let p'_rev = List.rev p' in
	if List.hd p'_rev = "" then
	  List.rev(List.tl p'_rev)
	else
	  p'
    | _ ->
	failwith ("Illegal posix path found in mount table: " ^ p)


(* -- For Cygwin-1.5 mount tables -- *)

(*
let read_mounts root =
  let table_hkey = walk_reg_path root false mount_table_path in
  let entries = List.map fst (reg_enum_key table_hkey) in
  let table =
    List.map
      (fun entry ->
	 let entry_hkey = reg_open_key table_hkey entry in
	 let flags =
	   try reg_int "flags" (reg_query_value entry_hkey "flags")
	   with _ -> 0 in
	 let native =
	   try reg_string "native" (reg_query_value entry_hkey "native")
	   with _ -> "" in
	 let posix = split_posix_path (String.lowercase entry) in
	 if List.mem "." posix || List.mem ".." posix then
	   failwith ("Illegal posix path found in mount table: " ^ entry);
	 reg_close_key entry_hkey;
	 { me_posix = posix;
	   me_win32 = (if native = "" then None else Some (slashify native));
	   me_flags = read_mount_flags flags
	 }
      )
      entries in
  reg_close_key table_hkey;
  table
 *)

let lines_re = Str.regexp "[\r\n]+"
let words_re = Str.regexp "[ \t]+"
let comma_re = Str.regexp ",+"
let bs_re = Str.regexp "\\\\[0-7][0-7][0-7]"

let decode_bs s =
  Str.global_substitute
    bs_re
    (fun _ ->
       let u = Str.matched_string s in
       let oct = String.copy "0o000" in
       oct.[2] <- u.[1];
       oct.[3] <- u.[2];
       oct.[4] <- u.[3];
       let n = int_of_string oct in
       let c = Char.chr n in
       String.make 1 c
    )
    s


let decode_flags s =
  let l = String.length s in
  let words =
    if l >= 2 && s.[0] = '(' && s.[l-1] = ')' then
      Str.split comma_re (String.sub s 1 (l-2))
    else
      [] in
  { mount_binary = List.mem "binary" words;
    mount_exec = List.mem "exec" words;
    mount_cygdrive = false;   (* not indicated *)
    mount_cygwin_exec = List.mem "cygexec" words;
    mount_notexec = List.mem "notexec" words;
    mount_acl = not (List.mem "noacl" words);
    mount_bind = List.mem "bind" words;
    mount_dos = List.mem "dos" words;
    mount_ihash = List.mem "ihash" words;
    mount_user = List.mem "user" words;
    mount_posix = not(List.mem "posix=0" words);
  }
  


let read_mounts2 s =
  let lines = Str.split lines_re s in
  List.flatten
    ( List.map
        (fun line ->
           let words = Str.split words_re line in
           try
             match words with
               | [ native; "on"; posix_s; "type"; _; flags_s; ] ->
	           let posix = 
                     split_posix_path (String.lowercase (decode_bs posix_s)) in
	           if List.mem "." posix || List.mem ".." posix then
	             failwith ("Illegal posix path found in mount table: " ^ 
                                 posix_s);
	           [ { me_posix = posix;
	               me_win32 = (if native = "" then None 
                                   else Some (decode_bs native));
                       me_flags = decode_flags flags_s
                     }
                   ]
               | _ ->
                   raise Not_found
           with
             | _ -> []
        )
        lines
    )


let mount_data = ref None


let get_mount_table () =
  let entries =
    try
      let filename = Sys.getenv "GODI_MOUNTS" in
      let f = open_in_bin filename in
      let size = in_channel_length f in
      let data = String.create size in
      really_input f data 0 size;
      close_in f;
      read_mounts2 data
    with
      | Not_found ->
          (* Run "mount" and save the result *)
          let b = Buffer.create 128 in
          let data = String.create 1024 in
          let ch = Unix.open_process_in "mount" in
          try
            while true do
              let n = input ch data 0 (String.length data) in
              if n=0 then raise End_of_file;
              Buffer.add_substring b data 0 n;
            done;
            assert false
          with 
            | End_of_file ->
                close_in ch;
                let data = Buffer.contents b in
                let entries = read_mounts2 data in
                mount_data := Some data;
                entries
  in
  { mt_entries = entries;
    mt_cygdrive_prefix = "/cygdrive";
    mt_cygdrive_flags = decode_flags ""
  }


let get_drives() =
  (* Return list with drive letters in lowercase *)
  let drives = get_logical_drives() in
  let l = ref [] in
  for k = 25 downto 0 do
    if drives land (1 lsl k) <> 0 then
      l := (String.make 1 (Char.chr (97+k))) :: !l
  done;
  !l



(* -- For Cygwin-1.5 mount tables -- *)

(*
let get_cygdrive_info root =
  (* Returns the /cygdrive prefix, and the flags *)
  let table_hkey = walk_reg_path root false mount_table_path in
  let prefix = 
    reg_string "cygdrive prefix"
      (reg_query_value table_hkey "cygdrive prefix") in
  let flags = 
    reg_int "cygdrive flags"
      (reg_query_value table_hkey "cygdrive flags") in
  (prefix, flags)


let get_mount_table() =
  (* Returns the merged mount table, i.e. after mixing user and system
     mounts. Missing entries for drives are added. Entries for "/proc"
     and "/dev" are also added, but without win32 path. The entries are
     returned in an order so that more specific entries come first.
     The POSIX names are returned in lowercase letters.
   *)
  let merged_table = Hashtbl.create 10 in
  let user_mounts = read_mounts hkey_current_user in
  let system_mounts = read_mounts hkey_local_machine in
  List.iter
    (fun me -> Hashtbl.replace merged_table me.me_posix me)
    system_mounts;
  List.iter
    (fun me -> Hashtbl.replace merged_table me.me_posix me)
    user_mounts;
  if not (Hashtbl.mem merged_table ["dev"]) then
    Hashtbl.add 
      merged_table
      ["dev"]
      { me_posix = ["dev"]; 
	me_win32 = None; 
	me_flags = read_mount_flags 0x020a
      };
  if not (Hashtbl.mem merged_table ["proc"]) then
    Hashtbl.add 
      merged_table
      ["proc"]
      { me_posix = ["proc"]; 
	me_win32 = None; 
	me_flags = read_mount_flags 0x040a
      };
  let cygdrive_prefix, cygdrive_flags = 
    try get_cygdrive_info hkey_current_user
    with _ -> get_cygdrive_info hkey_local_machine in
  let cygdrives = ref [] in
  List.iter
    (fun drive ->
       let posix_path = cygdrive_prefix ^ "/" ^ drive in
       let posix = split_posix_path posix_path in
       let me =
	 { me_posix = posix;
	   me_win32 = Some(drive ^ ":/");
	   me_flags = read_mount_flags cygdrive_flags
	 } in
       if not(Hashtbl.mem merged_table posix) then
	 cygdrives := me :: !cygdrives
    )
    (get_drives());
  let l = Hashtbl.fold (fun _ me acc -> me :: acc) merged_table [] in
  let l' = List.map (fun me -> (String.concat "/" me.me_posix, me)) l in
  let l'' =  List.sort (fun (x,_) (y,_) -> compare y x) l' in
  let entries = List.map snd l'' @ !cygdrives in
  { mt_entries = entries;
    mt_cygdrive_prefix = cygdrive_prefix;
    mt_cygdrive_flags = read_mount_flags cygdrive_flags;
  }
 *)


type split_filename =
  [ `Relative of string list
  | `POSIX_absolute of string list
  | `WIN32_drive of string * string list
  | `WIN32_remote of string * string list
  | `Error
  ]


let drive_re = Str.regexp "[a-zA-Z]:"
let unc_re = Str.regexp "//\\([^/]+\\)\\(/.*\\)$"
let pathsep_re = Str.regexp "/" 


let split mk_abs mk_rel empty s =
  let l = Str.split_delim pathsep_re s in
  match l with
    | [] ->
	empty
    | "" :: path ->
        mk_abs (List.filter (fun s -> s <> "") path)
    | path ->
        mk_rel (List.filter (fun s -> s <> "") path)



let get_splitname slashname =
  if Str.string_match drive_re slashname 0 then (
    if String.length slashname < 3 then
      `WIN32_drive(String.sub slashname 0 1, [])
    else
      split 
	(fun l -> `WIN32_drive(String.sub slashname 0 1, l))
	(fun _ ->
	   raise(Sys_error("Paths with drive letters must be absolute")))
	`Error
	(String.sub slashname 2 (String.length slashname - 2))
  )
  else
    if Str.string_match unc_re slashname 0 then (
      let host = Str.matched_group 1 slashname in
      let path = Str.matched_group 2 slashname in
      split
	(fun l -> `WIN32_remote(host,l))
	(fun _ -> assert false)
	`Error
	path
    )
    else
      split
	(fun l -> `POSIX_absolute l)
	(fun l -> `Relative l)
	`Error
	slashname


let rec is_prefix p1 p2 =
  match p1 with
    | [] -> true
    | n1 :: p1' ->
	( match p2 with
	    | [] -> false
	    | n2 :: p2' -> n1 = n2 && is_prefix p1' p2'
	)


let rec after_prefix p1 p2 =
  match p1 with
    | [] -> p2
    | n1 :: p1' ->
	( match p2 with
	    | [] -> invalid_arg "after_prefix"
	    | n2 :: p2' -> 
		(* if n1 <> n2 then  invalid_arg "after_prefix"; *)
		after_prefix p1' p2'
	)


let is_win32_prefix me splitname =
  match me.me_win32 with
    | Some win32 ->
	let me_splitname = get_splitname(slashify win32) in
	( match (me_splitname, splitname) with
	    | `WIN32_drive(d1,p1), `WIN32_drive(d2,p2) ->
		(String.lowercase d1 = String.lowercase d2) && 
	          (is_prefix 
		     (List.map String.lowercase p1)
		     (List.map String.lowercase p2))
	    | `WIN32_remote(h1,p1), `WIN32_remote(h2,p2) ->
		h1 = h2 && 
	          (is_prefix 
		     (List.map String.lowercase p1)
		     (List.map String.lowercase p2))
	    | _ ->
		false
	)

    | None ->
	false


let after_win32_prefix me splitname =
  match me.me_win32 with
    | Some win32 ->
	let me_splitname = get_splitname(slashify win32) in
	( match (me_splitname, splitname) with
	    | `WIN32_drive(d1,p1), `WIN32_drive(d2,p2) ->
		after_prefix p1 p2
	    | `WIN32_remote(h1,p1), `WIN32_remote(h2,p2) ->
		after_prefix p1 p2
	    | _ ->
		[]
	)

    | None ->
	[]



let rec translate_to_posix (mtable : mount_table)
                           (splitname : split_filename) =
  (* Translate splitname to a POSIX path, or return None if this is not
     possible (i.e. the name points to outside the mounted POSIX tree).
     Also return the effective mount flags.

     This function works only for win32 paths not containing "." or "..".
   *)
  match splitname with
    | `WIN32_drive(_,_)
    | `WIN32_remote(_,_) ->
	(* Look into the mount table: *)
	( try
	    let me =
	      List.find 
		(fun me -> is_win32_prefix me splitname)
		mtable.mt_entries in
	    let d = after_win32_prefix me splitname in
	    (Some(me.me_posix @ d, me.me_flags))
	  with Not_found ->
	    None
	)

    | _ ->
	invalid_arg "Godi_file_helper.translate_to_posix"


let slash_concat p1 p2 =
  if p1 <> "" && p1.[ String.length p1 - 1 ] = '/' then
    p1 ^ p2
  else
    p1 ^ "/" ^ p2
	


let translate_to_win32 (mtable : mount_table)
                       (splitname : split_filename)
		       (resolve_final_symlink : bool) =

  (* resolve_final_symlink: if the last component of the path is a symlink,
     it is only resolved if this argument is true
   *)

  let symlink_counter = ref 0 in

  let rec walk cwd_posix_opt cwd_win32 (cwd_flags :mount_flags option) p =
    (* cwd_posix_opt: the physical POSIX path (in lowercase). It is None if
                  we are outside the POSIX namespace.
       cwd_win32: the corresponding Win32 path (in original case, using 
                  slashes)
       cwd_flags: the flags from the mount point that counts, or None
       p: the path to resolve relative to cwd_posix/cwd_win32
     *)
    match p with
      | [] -> 
	  (* Corner case *)
	  (cwd_win32, cwd_flags)
      | [ "." ] ->
	  (* Keep dot in the trailing position. Remember that "foo/." only
             exists if it is a directory.
	   *)
	  (slash_concat cwd_win32 ".", cwd_flags)
      | "." :: p' ->
	  (* Other dots can be safely removed. *)
	  walk cwd_posix_opt cwd_win32 cwd_flags p'
      | ".." :: p' ->
	  ( match cwd_posix_opt with
	      | Some cwd_posix ->
		  (* Remove the last component from cwd_posix, and re-resolve
                     it *)
		  let new_posix =
		    if cwd_posix = [] then
		      []
		    else
		      List.rev (List.tl (List.rev cwd_posix)) in
		  walk_posix_absolute (new_posix @ ["."] @ p')
		    (* We insert the dot to ensure that new_posix is 
                       a directory *)
		    (* TODO: This is not fully correct. We should handle
                       the case that the mount point does not change
                       separately
		     *)
	      | None ->
		  (* Simply add ".." to the win32 path *)
		  walk None (slash_concat cwd_win32 "..") cwd_flags p'
	  )
      | name :: p' ->
	  (* If there is a mount table entry for cwd_posix/name, we have
             to take it into account.
	   *)
	  let name_lc = String.lowercase name in
	  let me_opt =
	    match cwd_posix_opt with
	      | Some cwd_posix ->
		  let next_posix = cwd_posix @ [name_lc] in
		  ( try
		      Some(List.find 
			     (fun me -> me.me_posix = next_posix)
			     mtable.mt_entries,
			   next_posix
			  )
		    with Not_found -> None
		  )
	      | None -> None in
	  ( match me_opt with
	      | Some(me,next_posix) ->
		  let next_win32 =
		    match me.me_win32 with
		      | Some p ->
                          if me.me_flags.mount_bind then
                            raise(Sys_error "Bind mounts not implemented");
                          p
		      | None ->
			  raise(Sys_error("No win32 path in mount table: /" ^ 
					    String.concat "/" next_posix)) in
		  walk (Some next_posix) next_win32 (Some me.me_flags) p'
	      | None ->
		  (* next, cwd_win32/name could be a symlink *)
		  let enable_symlink_resolver =
		    p' <> [] || resolve_final_symlink in
		  let sym_win32 = slash_concat cwd_win32 name in
		  ( match read_symlink sym_win32 with
		      | Some(target,sym_win32_full) when enable_symlink_resolver ->
			  incr symlink_counter;
			  if !symlink_counter > 50 then
			    raise(Sys_error 
				    ("Too many symlinks: " ^ sym_win32_full));
			  let split_target = get_splitname target in
			  ( match split_target with
			      | `POSIX_absolute q ->
				  walk_posix_absolute (q @ p')
			      | `WIN32_drive(drive,q) ->
				  walk_win32_drive drive (q @ p')
			      | `WIN32_remote(host,q) ->
				  walk_win32_remote host (q @ p')
			      | `Relative q ->
				  let symp = q @ p' in
				  walk cwd_posix_opt cwd_win32 cwd_flags symp
			      | `Error ->
				  raise(Sys_error ("Bad symlink: " ^
						     sym_win32_full))
			  )
		      | Some(_,sym_win32_full) ->
			  (* take the .lnk suffix into account *)
			  (sym_win32_full,
			   cwd_flags)
		      | None ->
			  (* hey, alright, this is a normal name *)
			  let next_posix_opt =
			    match cwd_posix_opt with
			      | Some cwd_posix -> Some(cwd_posix @ [name_lc])
			      | None -> None in
			  walk 
			    next_posix_opt
			    (slash_concat cwd_win32 name)
			    cwd_flags
			    p'
		  )
	  )

  and walk_posix_absolute p =
    (* Look into the mount table: *)
    let me =
      try
	List.find 
	  (fun me -> is_prefix me.me_posix p)
	  mtable.mt_entries
      with Not_found ->
	raise(Sys_error("Not found in mount table: /" ^ 
			  String.concat "/" p)) in
    let w32dir =
      match me.me_win32 with
	| Some p -> p
	| None ->
	    raise(Sys_error("No win32 path in mount table: /" ^ 
			      String.concat "/" p)) in
    let posix_rest = after_prefix me.me_posix p in
    walk (Some me.me_posix) w32dir (Some me.me_flags) posix_rest

  and walk_win32_drive drive p =
    (* Check if there is a mount point for the drive. If so, take the flags
       from there.
     *)
    let cygdrive = 
      mtable.mt_cygdrive_prefix ^ "/" ^ (String.lowercase drive) in
    let cygdrive_p =
      split 
	(fun l -> List.map String.lowercase l)
	(fun _ -> raise(Sys_error "Bad cygdrive prefix")) 
	[]
	cygdrive in
    let drive_flags =
      try
	let me = 
	  List.find
	    (fun me -> me.me_posix = cygdrive_p)
	    mtable.mt_entries in
	Some me.me_flags
      with Not_found -> None in
    let w32path = drive ^ ":/" in
    walk None w32path drive_flags p
    
  and walk_win32_remote host p =
    (* This is taken as entirely outside the POSIX hierarchy *)
    let w32path = "//" ^ host ^ "/" in
    walk None w32path None p
  in

  match splitname with
    | `WIN32_drive(drive,p) ->
	walk_win32_drive drive p
    | `WIN32_remote(host,p) ->
	walk_win32_remote host p
    | `POSIX_absolute p ->
	walk_posix_absolute p
    | `Relative p ->
	(* The Win32 working directory is
           translated to a POSIX path using the mount table
	 *)
	let cwd = slashify (Unix.getcwd()) in
	let cwd_win32 = get_splitname cwd in
	let posix_opt = translate_to_posix mtable cwd_win32 in
	let cwd_posix_opt, cwd_flags =
	  match posix_opt with
	    | Some(p,flags) -> (Some p), (Some flags)
	    | None -> (None, None) in
	walk cwd_posix_opt cwd cwd_flags p
    | `Error ->
	invalid_arg "Godi_file_helper.translate_to_win32"
